@@ -1,6 +1,6 @@
 /*****************************************************************************
  *  Written by Chris Dunlap <cdunlap@llnl.gov>.
- *  Copyright (C) 2007-2016 Lawrence Livermore National Security, LLC.
+ *  Copyright (C) 2007-2022 Lawrence Livermore National Security, LLC.
  *  Copyright (C) 2002-2007 The Regents of the University of California.
  *  UCRL-CODE-155910.
  *
@@ -219,8 +219,8 @@ dec_check_retry (munge_cred_t c)
 
     if (m->retry > 0) {
         log_msg (LOG_INFO,
-            "Decode retry #%d for client uid=%d gid=%d",
-            m->retry, m->client_uid, m->client_gid);
+            "Decode retry #%d for client UID=%u GID=%u", m->retry,
+            (unsigned int) m->client_uid, (unsigned int) m->client_gid);
     }
     if (m->retry > MUNGE_SOCKET_RETRY_ATTEMPTS) {
         return (m_msg_set_err (m, EMUNGE_SOCKET,
@@ -804,15 +804,21 @@ dec_unpack_inner (munge_cred_t c)
     /*
      *  Unpack the origin IP address.
      */
-    if (m->addr_len != sizeof (m->addr)) {
-        return (m_msg_set_err (m, EMUNGE_BAD_CRED,
-            strdup ("Invalid origin IP addr length")));
-    }
     if (m->addr_len > len) {
         return (m_msg_set_err (m, EMUNGE_BAD_CRED,
             strdup ("Truncated origin IP addr")));
     }
-    memcpy (&m->addr, p, m->addr_len);
+    else if (m->addr_len == 4) {
+        assert (sizeof (m->addr) == 4);
+        memcpy (&m->addr, p, m->addr_len);
+    }
+    else if (m->addr_len == 0) {
+        memset (&m->addr, 0, sizeof (m->addr));
+    }
+    else {
+        return (m_msg_set_err (m, EMUNGE_BAD_CRED,
+            strdup ("Invalid origin IP addr length")));
+    }
     p += m->addr_len;
     len -= m->addr_len;
     /*
@@ -954,8 +960,8 @@ dec_validate_auth (munge_cred_t c)
 
 unauthorized:
     return (m_msg_set_err (m, EMUNGE_CRED_UNAUTHORIZED,
-        strdupf ("Unauthorized credential for client UID=%d GID=%d",
-            m->client_uid, m->client_gid)));
+        strdupf ("Unauthorized credential for client UID=%u GID=%u",
+            (unsigned int) m->client_uid, (unsigned int) m->client_gid)));
 }
 
 
@@ -1012,9 +1018,9 @@ dec_validate_replay (munge_cred_t c)
         if ((conf->got_socket_retry)
                 && (m->retry > 0)
                 && (m->retry <= MUNGE_SOCKET_RETRY_ATTEMPTS)) {
-            log_msg (LOG_NOTICE,
-                "Allowed credential replay for client UID=%d GID=%d",
-                m->client_uid, m->client_gid);
+            log_msg (LOG_INFO,
+                "Allowed credential replay for client UID=%u GID=%u",
+                (unsigned int) m->client_uid, (unsigned int) m->client_gid);
             return (0);
         }
         else {
