@@ -1,6 +1,6 @@
 /*****************************************************************************
  *  Written by Chris Dunlap <cdunlap@llnl.gov>.
- *  Copyright (C) 2007-2016 Lawrence Livermore National Security, LLC.
+ *  Copyright (C) 2007-2022 Lawrence Livermore National Security, LLC.
  *  Copyright (C) 2002-2007 The Regents of the University of California.
  *  UCRL-CODE-155910.
  *
@@ -48,20 +48,25 @@
 #include <sys/un.h>
 #include <unistd.h>
 #include "auth_recv.h"
+#include "cipher.h"
 #include "common.h"
 #include "conf.h"
 #include "crypto.h"
+#include "daemonpipe.h"
 #include "gids.h"
+#include "hash.h"
 #include "job.h"
+#include "lock.h"
 #include "log.h"
+#include "md.h"
 #include "missing.h"
 #include "munge_defs.h"
 #include "path.h"
-#include "posignal.h"
 #include "random.h"
 #include "replay.h"
 #include "str.h"
 #include "timer.h"
+#include "xsignal.h"
 
 
 /*****************************************************************************
@@ -69,18 +74,14 @@
  *****************************************************************************/
 
 static void disable_core_dumps (void);
-static int daemonize_init (char *progname);
-static void daemonize_fini (int fd);
+static void daemonize_init (char *progname, conf_t conf);
+static void daemonize_fini (void);
 static void open_logfile (const char *logfile, int priority, int got_force);
 static void handle_signals (void);
-static void hup_handler (int signum);
-static void exit_handler (int signum);
+static void sig_handler (int sig);
 static void write_pidfile (const char *pidfile, int got_force);
 static void lock_memory (void);
 static void sock_create (conf_t conf);
-static void sock_lock (conf_t conf);
-static int set_file_lock (int fd);
-static pid_t is_file_locked (int fd);
 static void sock_destroy (conf_t conf);
 
 
@@ -88,7 +89,8 @@ static void sock_destroy (conf_t conf);
  *  Global Variables
  *****************************************************************************/
 
-volatile sig_atomic_t done = 0;         /* global flag set true for exit     */
+volatile sig_atomic_t got_reconfig = 0;     /* signum if HUP received        */
+volatile sig_atomic_t got_terminate = 0;    /* signum if INT/TERM received   */
 
 
 /*****************************************************************************
@@ -98,7 +100,6 @@ volatile sig_atomic_t done = 0;         /* global flag set true for exit     */
 int
 main (int argc, char *argv[])
 {
-    int   fd = -1;
     char *log_identity = argv[0];
     int   log_priority = LOG_INFO;
     int   log_options = LOG_OPT_PRIORITY;
@@ -112,26 +113,30 @@ main (int argc, char *argv[])
     disable_core_dumps ();
     conf = create_conf ();
     parse_cmdline (conf, argc, argv);
+    process_conf (conf);
     auth_recv_init (conf->auth_server_dir, conf->auth_client_dir,
         conf->got_force);
 
     if (!conf->got_foreground) {
-        fd = daemonize_init (argv[0]);
+        daemonize_init (argv[0], conf);
         if (conf->got_syslog) {
-            log_open_file (NULL, NULL, 0, 0);
+            log_close_file ();
             log_open_syslog (log_identity, LOG_DAEMON);
         }
         else {
             open_logfile (conf->logfile_name, log_priority, conf->got_force);
         }
     }
+    log_msg (LOG_NOTICE, "Starting %s-%s daemon (pid %d)",
+        PACKAGE, VERSION, (int) getpid ());
     handle_signals ();
-    lookup_ip_addr (conf);
-    write_pidfile (conf->pidfile_name, conf->got_force);
+    write_origin_addr (conf);
     if (conf->got_mlockall) {
         lock_memory ();
     }
     crypto_init ();
+    cipher_init_subsystem ();
+    md_init_subsystem ();
     if (random_init (conf->seed_name) < 0) {
         if (conf->seed_name) {
             free (conf->seed_name);
@@ -143,25 +148,25 @@ main (int argc, char *argv[])
     replay_init ();
     timer_init ();
     sock_create (conf);
+    write_pidfile (conf->pidfile_name, conf->got_force);
 
     if (!conf->got_foreground) {
-        daemonize_fini (fd);
+        daemonize_fini ();
     }
-    log_msg (LOG_NOTICE, "Starting %s daemon (pid %d)",
-        META_ALIAS, (int) getpid ());
-
     job_accept (conf);
 
     sock_destroy (conf);
     timer_fini ();
     replay_fini ();
     gids_destroy (conf->gids);
+    hash_drop_memory ();
     random_fini (conf->seed_name);
     crypto_fini ();
-    destroy_conf (conf);
+    destroy_conf (conf, 1);
 
-    log_msg (LOG_NOTICE, "Stopping %s daemon (pid %d)",
-        META_ALIAS, (int) getpid ());
+    log_msg (LOG_NOTICE, "Stopping %s-%s daemon (pid %d)",
+        PACKAGE, VERSION, (int) getpid ());
+    log_close_all ();
 
     exit (EMUNGE_SUCCESS);
 }
@@ -186,68 +191,62 @@ disable_core_dumps (void)
 }
 
 
-static int
-daemonize_init (char *progname)
+static void
+daemonize_init (char *progname, conf_t conf)
 {
 /*  Begins the daemonization of the process.
  *  Despite the fact that this routine backgrounds the process, control
  *    will not be returned to the shell until daemonize_fini() is called.
- *  Returns an 'fd' to pass to daemonize_fini() to complete the daemonization.
  */
-    int         fds [2];
-    pid_t       pid;
-    int         n;
-    signed char priority;
-    char        ebuf [1024];
+    pid_t pid;
+    int   status;
+    int   priority;
+    char  buf [1024];
 
     /*  Clear file mode creation mask.
      */
     umask (0);
 
-    /*  Create pipe for IPC so parent process will wait to terminate until
-     *    signaled by grandchild process.  This allows messages written to
-     *    stdout/stderr by the grandchild to be properly displayed before
-     *    the parent process returns control to the shell.
+    /*  Create a daemonpipe to have the parent process wait until signaled by
+     *    its double-forked grandchild process that startup is complete.
      */
-    if (pipe (fds) < 0) {
-        log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to create daemon pipe");
+    if (daemonpipe_create () < 0) {
+        log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to create daemonpipe");
     }
-    /*  Set the fd used by log_err() to return status back to the parent.
-     */
-    log_set_err_pipe (fds[1]);
-
     /*  Automatically background the process and
-     *    ensure child is not a process group leader.
+     *    ensure child process is not a process group leader.
      */
     if ((pid = fork ()) < 0) {
-        log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to create child process");
+        log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to fork child process");
     }
     else if (pid > 0) {
-        log_set_err_pipe (-1);
-        if (close (fds[1]) < 0) {
+        /*
+         *  Parent process waits for notification that startup is complete
+         *    before exiting.
+         */
+        if (daemonpipe_close_writes () < 0) {
             log_errno (EMUNGE_SNAFU, LOG_ERR,
-                "Failed to close write-pipe in parent process");
+                "Failed to close write-end of daemonpipe");
         }
-        if ((n = read (fds[0], &priority, sizeof (priority))) < 0) {
+        if (daemonpipe_read (&status, &priority, buf, sizeof (buf)) < 0) {
             log_errno (EMUNGE_SNAFU, LOG_ERR,
-                "Failed to read status from grandchild process");
+                "Failed to read from daemonpipe");
         }
-        if ((n > 0) && (priority >= 0)) {
-            if ((n = read (fds[0], ebuf, sizeof (ebuf))) < 0) {
-                log_errno (EMUNGE_SNAFU, LOG_ERR,
-                    "Failed to read err msg from grandchild process");
-            }
-            if ((n > 0) && (ebuf[0] != '\0')) {
-                log_open_file (stderr, progname, priority, LOG_OPT_PRIORITY);
-                log_msg (priority, "%s", ebuf);
+        if (status != 0) {
+            if ((priority >= 0) && (buf[0] != '\0')) {
+                log_msg (priority, "%s", buf);
             }
             exit (EXIT_FAILURE);
         }
+        destroy_conf (conf, 0);
+        log_close_all ();
         exit (EXIT_SUCCESS);
     }
-    if (close (fds[0]) < 0) {
+    /*  Child process continues.
+     */
+    if (daemonpipe_close_reads () < 0) {
         log_errno (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to close read-pipe in child process");
+            "Failed to close read-end of daemonpipe");
     }
     /*  Become a session leader and process group leader
      *    with no controlling tty.
@@ -256,31 +255,32 @@ daemonize_init (char *progname)
         log_errno (EMUNGE_SNAFU, LOG_ERR,
             "Failed to disassociate controlling tty");
     }
-    /*  Ignore SIGHUP to keep child from terminating when
-     *    the session leader (ie, the parent) terminates.
+    /*  Ignore SIGHUP to keep child process from terminating when
+     *    the session leader (i.e., the parent proces) terminates.
      */
-    if (posignal (SIGHUP, SIG_IGN) == SIG_ERR) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Failed to ignore signal=%d", SIGHUP);
-    }
-    /*  Abdicate session leader position in order to guarantee
-     *    daemon cannot automatically re-acquire a controlling tty.
+    xsignal_ignore (SIGHUP);
+
+    /*  Abdicate session leader position to ensure the daemon cannot
+     *    automatically re-acquire a controlling tty.
      */
     if ((pid = fork ()) < 0) {
-        log_errno (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to create grandchild process");
+        log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to fork grandchild process");
     }
     else if (pid > 0) {
+        destroy_conf (conf, 0);
+        log_close_all ();
         exit (EXIT_SUCCESS);
     }
-    return (fds[1]);
+    /*  Grandchild process continues.
+     */
+    return;
 }
 
 
 static void
-daemonize_fini (int fd)
+daemonize_fini (void)
 {
-/*  Completes the daemonization of the process,
- *    where 'fd' is the file descriptor returned by daemonize_init().
+/*  Completes the daemonization of the process.
  */
     int dev_null;
 
@@ -289,7 +289,7 @@ daemonize_fini (int fd)
      */
     if (chdir ("/") < 0) {
         log_errno (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to change to root directory");
+            "Failed to change CWD to root directory");
     }
     /*  Discard data to/from stdin, stdout, and stderr.
      */
@@ -308,18 +308,21 @@ daemonize_fini (int fd)
         log_errno (EMUNGE_SNAFU, LOG_ERR,
             "Failed to dup \"/dev/null\" onto stderr");
     }
-    if (close (dev_null) < 0) {
+    if ((dev_null > STDERR_FILENO) && (close (dev_null)) < 0) {
         log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to close \"/dev/null\"");
     }
-    /*  Clear the fd used by log_err() to return status back to the parent.
+    /*  Signal parent process to exit now that startup is complete.
+     *  The daemonpipe_write() below is not strictly necessary since
+     *    daemonpipe_close_writes() closes the daemonpipe which will cause
+     *    daemonpipe_read() to read an EOF.
      */
-    log_set_err_pipe (-1);
-    /*
-     *  Signal grandparent process to terminate.
-     */
-    if ((fd >= 0) && (close (fd) < 0)) {
+    if (daemonpipe_write (0, 0, NULL) < 0) {
         log_errno (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to close write-pipe in grandchild process");
+            "Failed to signal parent process that startup is complete");
+    }
+    if (daemonpipe_close_writes () < 0) {
+        log_errno (EMUNGE_SNAFU, LOG_ERR,
+            "Failed to close write-end of daemonpipe");
     }
     return;
 }
@@ -328,86 +331,69 @@ daemonize_fini (int fd)
 static void
 open_logfile (const char *logfile, int priority, int got_force)
 {
-    int          got_symlink;
+    int          is_symlink;
+    int          is_missing;
     struct stat  st;
-    int          n;
+    int          rv;
     char         logdir [PATH_MAX];
     char         ebuf [1024];
     mode_t       mask;
     FILE        *fp;
 
-    /*  Check file permissions and whatnot.
-     */
-    got_symlink = (lstat (logfile, &st) == 0) ? S_ISLNK (st.st_mode) : 0;
-
-    if (((n = stat (logfile, &st)) < 0) && (errno == ENOENT)) {
-        if (!got_symlink) {
-            ; /* A missing logfile is not considered an error. */
-        }
-        else if (!got_force) {
-            log_err (EMUNGE_SNAFU, LOG_ERR,
-                "Logfile is insecure: \"%s\" should be a regular file",
-                logfile);
-        }
-        else {
-            log_msg (LOG_WARNING,
-                "Logfile is insecure: \"%s\" should not be a symlink",
-                logfile);
-        }
+    if ((logfile == NULL) || (*logfile == '\0')) {
+        log_err (EMUNGE_SNAFU, LOG_ERR, "Logfile name is undefined");
     }
-    else {
-        if (n < 0) {
+    is_symlink = (lstat (logfile, &st) == 0) ? S_ISLNK (st.st_mode) : 0;
+    if (is_symlink) {
+        log_err_or_warn (got_force,
+            "Logfile is insecure: \"%s\" should not be a symbolic link",
+            logfile);
+    }
+    rv = stat (logfile, &st);
+    is_missing = (rv < 0) && (errno == ENOENT);
+
+    if (!is_missing) {
+        if (rv < 0) {
             log_errno (EMUNGE_SNAFU, LOG_ERR,
                 "Failed to check logfile \"%s\"", logfile);
         }
-        if (!S_ISREG (st.st_mode) || got_symlink) {
-            if (!got_force || !got_symlink)
-                log_err (EMUNGE_SNAFU, LOG_ERR,
-                    "Logfile is insecure: \"%s\" should be a regular file",
-                    logfile);
-            else
-                log_msg (LOG_WARNING,
-                    "Logfile is insecure: \"%s\" should not be a symlink",
-                    logfile);
+        if (!S_ISREG (st.st_mode)) {
+            log_err (EMUNGE_SNAFU, LOG_ERR,
+                "Logfile is insecure: \"%s\" must be a regular file "
+                "(type=%07o)", logfile, (st.st_mode & S_IFMT));
         }
         if (st.st_uid != geteuid ()) {
-            if (!got_force)
-                log_err (EMUNGE_SNAFU, LOG_ERR,
-                    "Logfile is insecure: \"%s\" should be owned by UID %u",
-                    logfile, (unsigned) geteuid ());
-            else
-                log_msg (LOG_WARNING,
-                    "Logfile is insecure: \"%s\" should be owned by UID %u",
-                    logfile, (unsigned) geteuid ());
+            log_err_or_warn (got_force,
+                "Logfile is insecure: \"%s\" should be owned by UID %u "
+                "instead of UID %u", logfile, (unsigned) geteuid (),
+                (unsigned) st.st_uid);
         }
-        if (st.st_mode & (S_IWGRP | S_IWOTH)) {
-            if (!got_force)
-                log_err (EMUNGE_SNAFU, LOG_ERR,
-                    "Logfile is insecure: \"%s\" should not be writable "
-                    "by group or world", logfile);
-            else
-                log_msg (LOG_WARNING,
-                    "Logfile is insecure: \"%s\" should not be writable "
-                    "by group or world", logfile);
+        if (st.st_mode & S_IWGRP) {
+            log_err_or_warn (got_force,
+                "Logfile is insecure: \"%s\" should not be writable by group "
+                "(perms=%04o)", logfile, (st.st_mode & ~S_IFMT));
+        }
+        if (st.st_mode & S_IWOTH) {
+            log_err_or_warn (got_force,
+                "Logfile is insecure: \"%s\" should not be writable by other "
+                "(perms=%04o)", logfile, (st.st_mode & ~S_IFMT));
         }
     }
     /*  Ensure logfile dir is secure against modification by others.
      */
-    if (path_dirname (logfile, logdir, sizeof (logdir)) < 0) {
+    rv = path_dirname (logfile, logdir, sizeof (logdir));
+    if (rv < 0) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
             "Failed to determine dirname of logfile \"%s\"", logfile);
     }
-    n = path_is_secure (logdir, ebuf, sizeof (ebuf),
+    rv = path_is_secure (logdir, ebuf, sizeof (ebuf),
         PATH_SECURITY_IGNORE_GROUP_WRITE);
-    if (n < 0) {
+    if (rv < 0) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
             "Failed to check logfile dir \"%s\": %s", logdir, ebuf);
     }
-    else if ((n == 0) && (!got_force)) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Logfile is insecure: %s", ebuf);
-    }
-    else if (n == 0) {
-        log_msg (LOG_WARNING, "Logfile is insecure: %s", ebuf);
+    else if (rv == 0) {
+        log_err_or_warn (got_force, "Logfile is insecure: %s", ebuf);
     }
     /*  Protect logfile against unauthorized access by removing write-access
      *    from group and all access from other.
@@ -430,37 +416,51 @@ open_logfile (const char *logfile, int priority, int got_force)
 static void
 handle_signals (void)
 {
-    if (posignal (SIGHUP, hup_handler) == SIG_ERR) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Failed to handle signal=%d", SIGHUP);
+    struct sigaction sa;
+    int              sig;
+    int              rv;
+
+    sa.sa_handler = sig_handler;
+    sa.sa_flags = 0;
+    rv = sigfillset (&sa.sa_mask);
+    if (rv == -1) {
+        log_errno (EMUNGE_SNAFU, LOG_ERR,
+                "Failed to initialize signal set to full");
     }
-    if (posignal (SIGINT, exit_handler) == SIG_ERR) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Failed to handle signal=%d", SIGINT);
+    sig = SIGHUP;
+    rv = sigaction (sig, &sa, NULL);
+    if (rv == -1) {
+        log_errno (EMUNGE_SNAFU, LOG_ERR,
+                "Failed to set handler for signal %d (%s)", sig,
+                strsignal (sig));
     }
-    if (posignal (SIGTERM, exit_handler) == SIG_ERR) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Failed to handle signal=%d", SIGTERM);
+    sig = SIGINT;
+    rv = sigaction (sig, &sa, NULL);
+    if (rv == -1) {
+        log_errno (EMUNGE_SNAFU, LOG_ERR,
+                "Failed to set handler for signal %d (%s)", sig,
+                strsignal (sig));
     }
-    if (posignal (SIGPIPE, SIG_IGN) == SIG_ERR) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Failed to ignore signal=%d", SIGPIPE);
+    sig = SIGTERM;
+    rv = sigaction (sig, &sa, NULL);
+    if (rv == -1) {
+        log_errno (EMUNGE_SNAFU, LOG_ERR,
+                "Failed to set handler for signal %d (%s)", sig,
+                strsignal (sig));
     }
+    xsignal_ignore (SIGPIPE);
     return;
 }
 
 
 static void
-hup_handler (int signum)
+sig_handler (int sig)
 {
-    if (conf) {
-        gids_update (conf->gids);
+    if (sig == SIGHUP) {
+        got_reconfig = sig;
     }
-    return;
-}
-
-
-static void
-exit_handler (int signum)
-{
-    if (!done) {
-        done = signum;
+    else if ((sig == SIGINT) || (sig == SIGTERM)) {
+        got_terminate = sig;
     }
     return;
 }
@@ -471,68 +471,85 @@ write_pidfile (const char *pidfile, int got_force)
 {
 /*  Creates the specified pidfile.
  *  The pidfile must be created after the daemon has finished forking.
+ *    It should be written after validation checks that might prevent the
+ *    daemon from starting (e.g., after creating the socket and obtaining
+ *    the lock), but before the original parent process terminates (i.e.,
+ *    before daemonize_fini()).
  */
     char    piddir [PATH_MAX];
     char    ebuf [1024];
-    int     n;
+    int     rv;
     mode_t  mask;
     FILE   *fp;
 
-    assert (pidfile != NULL);
-
+    if ((pidfile == NULL) || (*pidfile == '\0')) {
+        log_err (EMUNGE_SNAFU, LOG_ERR, "PIDfile name is undefined");
+    }
     /*  The pidfile must be specified with an absolute pathname; o/w, the
      *    unlink() call in destroy_conf() will fail because the daemon has
      *    chdir()'d.
      */
     if (pidfile[0] != '/') {
         log_err (EMUNGE_SNAFU, LOG_ERR,
-            "Pidfile \"%s\" requires an absolute path", pidfile);
+            "PIDfile \"%s\" requires an absolute path", pidfile);
     }
     /*  Ensure pidfile dir is secure against modification by others.
      */
     if (path_dirname (pidfile, piddir, sizeof (piddir)) < 0) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to determine dirname of pidfile \"%s\"", pidfile);
+            "Failed to determine dirname of PIDfile \"%s\"", pidfile);
     }
-    n = path_is_secure (piddir, ebuf, sizeof (ebuf), PATH_SECURITY_NO_FLAGS);
-    if (n < 0) {
+    rv = path_is_secure (piddir, ebuf, sizeof (ebuf), PATH_SECURITY_NO_FLAGS);
+    if (rv < 0) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to check pidfile dir \"%s\": %s", piddir, ebuf);
+            "Failed to check PIDfile dir \"%s\": %s", piddir, ebuf);
     }
-    else if ((n == 0) && (!got_force)) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Pidfile is insecure: %s", ebuf);
-    }
-    else if (n == 0) {
-        log_msg (LOG_WARNING, "Pidfile is insecure: %s", ebuf);
+    else if (rv == 0) {
+        log_err_or_warn (got_force, "PIDfile is insecure: %s", ebuf);
     }
     /*  Protect pidfile against unauthorized access by removing write-access
      *    from group and other.
+     *  An error removing an old pidfile is not considered fatal.
      */
     mask = umask (0);
     umask (mask | 022);
-    (void) unlink (pidfile);
+    do {
+        rv = unlink (pidfile);
+    } while ((rv < 0) && (errno == EINTR));
+
+    if ((rv < 0) && (errno != ENOENT)) {
+        log_msg (LOG_WARNING, "Failed to remove PIDfile \"%s\": %s",
+                pidfile, strerror (errno));
+    }
     fp = fopen (pidfile, "w");
     umask (mask);
     /*
      *  An error in creating the pidfile is not considered fatal.
      */
     if (!fp) {
-        log_msg (LOG_WARNING, "Failed to open pidfile \"%s\": %s",
+        log_msg (LOG_WARNING, "Failed to open PIDfile \"%s\": %s",
             pidfile, strerror (errno));
     }
     else if (fprintf (fp, "%d\n", (int) getpid ()) == EOF) {
-        log_msg (LOG_WARNING, "Failed to write to pidfile \"%s\": %s",
+        log_msg (LOG_WARNING, "Failed to write to PIDfile \"%s\": %s",
             pidfile, strerror (errno));
         (void) fclose (fp);
     }
     else if (fclose (fp) == EOF) {
-        log_msg (LOG_WARNING, "Failed to close pidfile \"%s\": %s",
+        log_msg (LOG_WARNING, "Failed to close PIDfile \"%s\": %s",
             pidfile, strerror (errno));
     }
     else {
         return;                         /* success */
     }
-    (void) unlink (pidfile);
+    do {
+        rv = unlink (pidfile);
+    } while ((rv < 0) && (errno == EINTR));
+
+    if ((rv < 0) && (errno != ENOENT)) {
+        log_msg (LOG_WARNING, "Failed to remove PIDfile \"%s\": %s",
+                pidfile, strerror (errno));
+    }
     return;                             /* failure */
 }
 
@@ -575,11 +592,11 @@ sock_create (conf_t conf)
 {
     char                sockdir [PATH_MAX];
     char                ebuf [1024];
-    int                 n;
     int                 sd;
     struct sockaddr_un  addr;
     mode_t              mask;
     int                 rv;
+    size_t              n;
 
     assert (conf != NULL);
 
@@ -588,41 +605,51 @@ sock_create (conf_t conf)
     }
     /*  Ensure socket dir is secure against modification by others.
      */
-    if (path_dirname (conf->socket_name, sockdir, sizeof (sockdir)) < 0) {
+    rv = path_dirname (conf->socket_name, sockdir, sizeof (sockdir));
+    if (rv < 0) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
             "Failed to determine dirname of socket \"%s\"", conf->socket_name);
     }
-    n = path_is_secure (sockdir, ebuf, sizeof (ebuf), PATH_SECURITY_NO_FLAGS);
-    if (n < 0) {
+    rv = path_is_secure (sockdir, ebuf, sizeof (ebuf), PATH_SECURITY_NO_FLAGS);
+    if (rv < 0) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
             "Failed to check socket dir \"%s\": %s", sockdir, ebuf);
     }
-    else if ((n == 0) && (!conf->got_force)) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Socket is insecure: %s", ebuf);
-    }
-    else if (n == 0) {
-        log_msg (LOG_WARNING, "Socket is insecure: %s", ebuf);
+    else if (rv == 0) {
+        log_err_or_warn (conf->got_force, "Socket is insecure: %s", ebuf);
     }
     /*  Ensure socket dir is accessible by all.
      */
-    n = path_is_accessible (sockdir, ebuf, sizeof (ebuf));
-    if (n < 0) {
+    rv = path_is_accessible (sockdir, ebuf, sizeof (ebuf));
+    if (rv < 0) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
             "Failed to check socket dir \"%s\": %s", sockdir, ebuf);
     }
-    else if ((n == 0) && (!conf->got_force)) {
-        log_err (EMUNGE_SNAFU, LOG_ERR, "Socket is inaccessible: %s", ebuf);
-    }
-    else if (n == 0) {
-        log_msg (LOG_WARNING, "Socket is inaccessible: %s", ebuf);
+    else if (rv == 0) {
+        log_err_or_warn (conf->got_force, "Socket is inaccessible: %s", ebuf);
     }
     /*  Create lockfile for exclusive access to the socket.
      */
-    sock_lock (conf);
+    lock_create (conf);
     /*
-     *  Create socket for communicating with clients.
+     *  Remove existing socket from previous instance.
      */
-    if ((sd = socket (PF_UNIX, SOCK_STREAM, 0)) < 0) {
+    do {
+        rv = unlink (conf->socket_name);
+    } while ((rv < 0) && (errno == EINTR));
+
+    if ((rv < 0) && (errno != ENOENT)) {
+        log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to remove socket \"%s\"",
+            conf->socket_name);
+    }
+    else if (rv == 0) {
+        log_msg (LOG_INFO, "Removed existing socket \"%s\"",
+            conf->socket_name);
+    }
+    /*  Create socket for communicating with clients.
+     */
+    sd = socket (PF_UNIX, SOCK_STREAM, 0);
+    if (sd < 0) {
         log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to create socket");
     }
     memset (&addr, 0, sizeof (addr));
@@ -630,7 +657,8 @@ sock_create (conf_t conf)
     n = strlcpy (addr.sun_path, conf->socket_name, sizeof (addr.sun_path));
     if (n >= sizeof (addr.sun_path)) {
         log_err (EMUNGE_SNAFU, LOG_ERR,
-            "Exceeded maximum length of socket pathname");
+            "Exceeded maximum length of %lu bytes for socket pathname",
+            sizeof (addr.sun_path));
     }
     /*  Ensure socket is accessible by all.
      */
@@ -640,199 +668,60 @@ sock_create (conf_t conf)
 
     if (rv < 0) {
         log_errno (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to bind \"%s\"", conf->socket_name);
+            "Failed to bind socket \"%s\"", conf->socket_name);
     }
     if (listen (sd, MUNGE_SOCKET_BACKLOG) < 0) {
         log_errno (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to listen on \"%s\"", conf->socket_name);
+            "Failed to listen on socket \"%s\"", conf->socket_name);
     }
     conf->ld = sd;
+    log_msg (LOG_INFO, "Created socket \"%s\"", conf->socket_name);
     return;
 }
 
 
-static void
-sock_lock (conf_t conf)
-{
-/*  Ensures exclusive access to the unix domain socket.
- */
-    struct stat  st;
-    mode_t       mask;
-    int          rv;
-
-    assert (conf != NULL);
-    assert (conf->lockfile_name == NULL);
-    assert (conf->lockfile_fd == -1);
-    assert (conf->socket_name != NULL);
-
-    if (!(conf->lockfile_name = strdupf ("%s.lock", conf->socket_name))) {
-        log_errno (EMUNGE_NO_MEMORY, LOG_ERR,
-            "Failed to create lockfile string");
-    }
-    if (conf->got_force) {
-        if ((unlink (conf->lockfile_name) < 0) && (errno != ENOENT)) {
-            log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to remove \"%s\"",
-                conf->lockfile_name);
-        }
-    }
-    else if (lstat (conf->lockfile_name, &st) < 0) {
-        if (errno != ENOENT) {
-            log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to stat \"%s\"",
-                conf->lockfile_name);
-        }
-    }
-    else if (!S_ISREG(st.st_mode)) {
-        log_err (EMUNGE_SNAFU, LOG_ERR,
-            "Lockfile is suspicious: \"%s\" should be a regular file",
-            conf->lockfile_name);
-    }
-    else if (st.st_uid != geteuid()) {
-        log_err (EMUNGE_SNAFU, LOG_ERR,
-            "Lockfile is suspicious: \"%s\" should be owned by UID %u",
-            conf->lockfile_name, (unsigned) geteuid());
-    }
-    else if ((st.st_mode & 07777) != S_IWUSR) {
-        log_err (EMUNGE_SNAFU, LOG_ERR,
-            "Lockfile is suspicious: \"%s\" should be writable only by user",
-            conf->lockfile_name);
-    }
-    mask = umask (0);
-    conf->lockfile_fd = creat (conf->lockfile_name, S_IWUSR);
-    umask (mask);
-
-    if (conf->lockfile_fd < 0) {
-        log_errno (EMUNGE_SNAFU, LOG_ERR,
-            "Failed to create \"%s\"", conf->lockfile_name);
-    }
-    if ((rv = set_file_lock (conf->lockfile_fd)) < 0) {
-        if (!conf->got_force)
-            log_errno (EMUNGE_SNAFU, LOG_ERR,
-                "Failed to lock \"%s\"", conf->lockfile_name);
-        else
-            log_msg (LOG_WARNING,
-                "Failed to lock \"%s\"", conf->lockfile_name);
-    }
-    else if (rv > 0) {
-
-        pid_t pid = is_file_locked (conf->lockfile_fd);
-
-        if (pid < 0) {
-            log_errno (EMUNGE_SNAFU, LOG_ERR,
-                "Failed to test lock \"%s\"", conf->lockfile_name);
-        }
-        else if (pid > 0) {
-            log_err (EMUNGE_SNAFU, LOG_ERR,
-                "Found pid %d bound to socket \"%s\"", pid, conf->socket_name);
-        }
-        else {
-            log_err (EMUNGE_SNAFU, LOG_ERR,
-                "Found inconsistent state for lock \"%s\"",
-                conf->lockfile_name);
-        }
-    }
-    if (unlink (conf->socket_name) == 0) {
-        log_msg (LOG_INFO, "Removed existing socket \"%s\"",
-            conf->socket_name);
-    }
-    else if (errno != ENOENT) {
-        log_errno (EMUNGE_SNAFU, LOG_ERR, "Failed to remove \"%s\"",
-            conf->socket_name);
-    }
-    return;
-}
-
-
-static int
-set_file_lock (int fd)
-{
-/*  Sets an exclusive advisory lock on the open file descriptor 'fd'.
- *  Returns 0 on success, 1 if a conflicting lock is held by another process,
- *    or -1 on error (with errno set).
- */
-    struct flock  fl;
-    int           rv;
-
-    if (fd < 0) {
-        errno = EBADF;
-        return (-1);
-    }
-    fl.l_type = F_WRLCK;
-    fl.l_whence = SEEK_SET;
-    fl.l_start = 0;
-    fl.l_len = 0;
-
-    rv = fcntl (fd, F_SETLK, &fl);
-    if (rv < 0) {
-        if ((errno == EACCES) || (errno == EAGAIN)) {
-            return (1);
-        }
-        return (-1);
-    }
-    return (0);
-}
-
-
-static pid_t
-is_file_locked (int fd)
-{
-/*  Tests whether an exclusive advisory lock could be obtained for the open
- *    file descriptor 'fd'.
- *  Returns 0 if the file is not locked, >0 for the pid of another process
- *    holding a conflicting lock, or -1 on error (with errno set).
- */
-    struct flock  fl;
-    int           rv;
-
-    if (fd < 0) {
-        errno = EBADF;
-        return (-1);
-    }
-    fl.l_type = F_WRLCK;
-    fl.l_whence = SEEK_SET;
-    fl.l_start = 0;
-    fl.l_len = 0;
-
-    rv = fcntl (fd, F_GETLK, &fl);
-    if (rv < 0) {
-        return (-1);
-    }
-    if (fl.l_type == F_UNLCK) {
-        return (0);
-    }
-    return (fl.l_pid);
-}
-
-
 static void
 sock_destroy (conf_t conf)
 {
+    int rv;
+
     assert (conf != NULL);
     assert (conf->ld >= 0);
     assert (conf->socket_name != NULL);
 
     if (conf->socket_name) {
-        if (unlink (conf->socket_name) < 0) {
-            log_msg (LOG_WARNING, "Failed to remove \"%s\": %s",
+        do {
+            rv = unlink (conf->socket_name);
+        } while ((rv < 0) && (errno == EINTR));
+
+        if (rv < 0) {
+            log_msg (LOG_WARNING, "Failed to remove socket \"%s\": %s",
                 conf->socket_name, strerror (errno));
         }
     }
     if (conf->ld >= 0) {
-        if (close (conf->ld) < 0) {
-            log_msg (LOG_WARNING, "Failed to close \"%s\": %s",
-                conf->ld, strerror (errno));
+        rv = close (conf->ld);
+        if (rv < 0) {
+            log_msg (LOG_WARNING, "Failed to close socket \"%s\": %s",
+                conf->socket_name, strerror (errno));
         }
         conf->ld = -1;
     }
     if (conf->lockfile_name) {
-        if (unlink (conf->lockfile_name) < 0) {
-            log_msg (LOG_WARNING, "Failed to remove \"%s\": %s",
+        do {
+            rv = unlink (conf->lockfile_name);
+        } while ((rv < 0) && (errno == EINTR));
+
+        if (rv < 0) {
+            log_msg (LOG_WARNING, "Failed to remove lockfile \"%s\": %s",
                 conf->lockfile_name, strerror (errno));
         }
     }
     if (conf->lockfile_fd >= 0) {
-        if (close (conf->lockfile_fd) < 0) {
-            log_msg (LOG_WARNING, "Failed to close \"%s\": %s",
-                conf->lockfile_fd, strerror (errno));
+        rv = close (conf->lockfile_fd);
+        if (rv < 0) {
+            log_msg (LOG_WARNING, "Failed to close lockfile \"%s\": %s",
+                conf->lockfile_name, strerror (errno));
         }
         conf->lockfile_fd = -1;
     }
